package proxy

import (
	"errors"
	"math/rand"
	"net/http"
	"time"

	"github.com/fagongzi/gateway/pkg/filter"
	"github.com/fagongzi/gateway/pkg/pb/metapb"
)

const (
	// RateBase base rate
	RateBase = 100
)

var (
	// ErrCircuitClose resource is in circuit close
	ErrCircuitClose = errors.New("resource is in circuit close")
	// ErrCircuitHalfLimited resource is in circuit half, traffic limit
	ErrCircuitHalfLimited = errors.New("resource is in circuit half, traffic limit")

	rd = rand.New(rand.NewSource(time.Now().UnixNano()))
)

// CircuitBreakeFilter CircuitBreakeFilter
type CircuitBreakeFilter struct {
	filter.BaseFilter
}

func newCircuitBreakeFilter() filter.Filter {
	return &CircuitBreakeFilter{}
}

// Init init filter
func (f *CircuitBreakeFilter) Init(cfg string) error {
	return nil
}

// Name return name of this filter
func (f *CircuitBreakeFilter) Name() string {
	return FilterCircuitBreake
}

// Pre execute before proxy
func (f *CircuitBreakeFilter) Pre(c filter.Context) (statusCode int, err error) {
	pc := c.(*proxyContext)
	cb := pc.circuitBreaker()
	if cb == nil {
		return f.BaseFilter.Pre(c)
	}

	protectedResourceStatus := pc.circuitStatus()
	protectedResource := pc.circuitResourceID()

	switch protectedResourceStatus {
	case metapb.Open:
		if c.Analysis().GetRecentlyRequestFailureRate(protectedResource, time.Duration(cb.RateCheckPeriod)) >= int(cb.FailureRateToClose) {
			pc.changeCircuitStatusToClose()
			c.Analysis().Reject(protectedResource)
			return http.StatusServiceUnavailable, ErrCircuitClose
		}

		return http.StatusOK, nil
	case metapb.Half:
		if limitAllow(cb.HalfTrafficRate) {
			return f.BaseFilter.Pre(c)
		}

		c.Analysis().Reject(protectedResource)
		return http.StatusServiceUnavailable, ErrCircuitHalfLimited
	default:
		c.Analysis().Reject(protectedResource)
		return http.StatusServiceUnavailable, ErrCircuitClose
	}
}

// Post execute after proxy
func (f *CircuitBreakeFilter) Post(c filter.Context) (statusCode int, err error) {
	pc := c.(*proxyContext)
	cb := pc.circuitBreaker()
	if cb == nil {
		return f.BaseFilter.Post(c)
	}

	protectedResourceStatus := pc.circuitStatus()
	protectedResource := pc.circuitResourceID()

	if protectedResourceStatus == metapb.Half &&
		c.Analysis().GetRecentlyRequestSuccessedRate(protectedResource, time.Duration(cb.RateCheckPeriod)) >= int(cb.SucceedRateToOpen) {
		pc.changeCircuitStatusToOpen()
	}

	return f.BaseFilter.Post(c)
}

// PostErr execute proxy has errors
func (f *CircuitBreakeFilter) PostErr(c filter.Context) {
	pc := c.(*proxyContext)
	cb := pc.circuitBreaker()
	if cb == nil {
		f.BaseFilter.PostErr(c)
		return
	}

	protectedResourceStatus := pc.circuitStatus()
	protectedResource := pc.circuitResourceID()

	if protectedResourceStatus == metapb.Half &&
		c.Analysis().GetRecentlyRequestFailureRate(protectedResource, time.Duration(cb.RateCheckPeriod)) >= int(cb.FailureRateToClose) {
		pc.changeCircuitStatusToClose()
	}
}

func limitAllow(rate int32) bool {
	randValue := rd.Intn(RateBase)
	return randValue < int(rate)
}
